@@ -637,8 +637,13 @@ void cookie_list(struct unix_ctx *ctx, SecmListCookiesReplyMsg *rep, FILE *out,
 		}
 
 		t = rep->cookies[i]->last_modified;
-		tm = localtime(&t);
-		strftime(str_since, sizeof(str_since), DATE_TIME_FMT, tm);
+		if (t > 0) {
+			tm = localtime(&t);
+			strftime(str_since, sizeof(str_since), DATE_TIME_FMT, tm);
+		} else {
+			str_since[0] = '?';
+			str_since[1] = 0;
+		}
 
 		groupname = rep->cookies[i]->groupname;
 		if (groupname == NULL || groupname[0] == 0)
@@ -748,7 +753,7 @@ int handle_list_iroutes_cmd(struct unix_ctx *ctx, const char *arg, cmd_params_st
 			}
 
 			for (j=0;j<rep->user[i]->n_iroutes;j++)
-				fprintf(out, "%6d %8s %6s %16s %28s",
+				fprintf(out, "%6d %8s %6s %16s %28s\n",
 					(int)rep->user[i]->id, username, rep->user[i]->tun, vpn_ip, rep->user[i]->iroutes[j]);
 
 		}
