@@ -46,7 +46,11 @@ fi
 
 update_config() {
 	file=$1
+	username=$(whoami)
+	group=$(groups|cut -f 1 -d ' ')
 	cp ${srcdir}/${file} "$file.tmp"
+	sed -i 's|@USERNAME@|'${username}'|g' "$file.tmp"
+	sed -i 's|@GROUP@|'${group}'|g' "$file.tmp"
 	sed -i 's|@SRCDIR@|'${srcdir}'|g' "$file.tmp"
 	CONFIG="$file.tmp"
 }
